@@ -1,4 +1,4 @@
-__version__ = "1.6.1"
+__version__ = "1.7.10"
 
 from deriva.core.utils.core_utils import *
 from deriva.core.base_cli import BaseCLI, KeyValuePairArgs
@@ -17,7 +17,8 @@ def get_credential(host,
                    config_file=DEFAULT_CONFIG_FILE,
                    requested_scope=None,
                    force_scope_lookup=False,
-                   match_scope_tag="deriva-all"):
+                   match_scope_tag="deriva-all",
+                   update_bdbag_keychain=True):
     """
     This function is used to get authorization credentials (in dict form) for use with various deriva-py API calls
     which take it as a parameter. A user must have already authenticated to the target host using either `deriva-auth`
@@ -36,7 +37,9 @@ def get_credential(host,
     :param match_scope_tag: In the case that a host-to-scope mapping request returns multiple scopes, this is the key
         value ("tag") to match against in the result dict. By convention, the default is set to "deriva-all", which is
         the expected response from webauthn.
-
+    :param update_bdbag_keychain: Updates the bdbag keychain file with the bearer token mapped to `host`. This is
+        done to ensure that the bdbag keychain is updated when a refreshable bearer-token gets refreshed during the
+        login check. Defaults to True.
     :return: A dict containing credential authorization values mapped by authorization type
     """
     # load deriva credential set first
@@ -44,13 +47,15 @@ def get_credential(host,
     creds = credentials.get(host, credentials.get(host.lower(), dict()))
 
     # load globus credentials and merge, if present
-    if os.path.isfile(globus_credential_file) and os.path.getsize(globus_credential_file) > 10:  # Don't load empty json
+    if globus_credential_file is not None and \
+            os.path.isfile(globus_credential_file) and \
+            os.path.getsize(globus_credential_file) > 10:  # Don't load empty json
         try:
             globus_client = GlobusNativeLogin(hosts=[host], config_file=config_file)
             scope_map = globus_client.hosts_to_scope_map(hosts=[host], match_scope_tag=match_scope_tag,
                                                          force_refresh=force_scope_lookup,
                                                          warn_on_discovery_failure=True if not creds else False)
-            tokens = globus_client.is_logged_in(exclude_defaults=True)
+            tokens = globus_client.is_logged_in(exclude_defaults=True, hosts_to_scope_map=scope_map)
             if tokens:
                 # 1. look for the explicitly requested scope in the token store, if specified
                 token = globus_client.find_access_token_for_scope(requested_scope, tokens)
@@ -62,6 +67,8 @@ def get_credential(host,
                                                                      match_scope_tag=match_scope_tag)
                 if token:
                     creds["bearer-token"] = token
+                    if update_bdbag_keychain:
+                        globus_client.update_bdbag_keychain(token=token, host=host)
         except Exception as e:
             logging.warning("Exception while getting Globus credentials: %s" % format_exception(e))
 
