@@ -7,9 +7,10 @@ import codecs
 import csv
 import json
 import requests
+from typing import NamedTuple
 
 from . import urlquote, urlsplit, urlunsplit, datapath, DEFAULT_HEADERS, DEFAULT_CHUNK_SIZE, DEFAULT_SESSION_CONFIG, \
-    Megabyte, Kilobyte, get_transfer_summary, IS_PY2
+    Megabyte, Kilobyte, get_transfer_summary
 from .deriva_binding import DerivaBinding, DerivaPathError
 from . import ermrest_model
 from .ermrest_model import nochange
@@ -52,11 +53,15 @@ class DerivaServer (DerivaBinding):
         """
         return ErmrestCatalog.connect(self, catalog_id, snaptime)
 
-    def create_ermrest_catalog(self, id=None, owner=None):
+    def create_ermrest_catalog(self, id=None, owner=None, name=None, description=None, is_persistent=None, clone_source=None):
         """Create an ERMrest catalog.
 
         :param id: The (str) id desired by the client (default None)
         :param owner: The initial (list of str) ACL desired by the client (default None)
+        :param name: Initial (str) catalog name if not None
+        :param description: Initial (str) catalog description if not None
+        :param is_persistent: Initial (bool) catalog persistence flag if not None
+        :param clone_source: Initial catalog clone_source if not None
 
         The new catalog id will be returned in the response, and used
         in future catalog access. The use of the id parameter
@@ -76,8 +81,17 @@ class DerivaServer (DerivaBinding):
         owner ACL influences which client(s) are allowed to retry
         creation with the same id.
 
+        The name, description, is_persistent, and clone_source
+        parameters are passed through to the catalog creation service
+        to initialize those respective metadata fields of the new
+        catalog's registry entry. See ERMrest documentation for more
+        detail. Authorization failures may occur when attempting to
+        set the is_persistent flag. By default, these fields are not
+        initialized in the catalog creation request, and they instead
+        receive server-assigned defaults.
+
         """
-        return ErmrestCatalog.create(self, id, owner)
+        return ErmrestCatalog.create(self, id, owner, name, description, is_persistent, clone_source)
 
     def connect_ermrest_alias(self, id):
         """Connect to an ERMrest alias and return the alias binding.
@@ -87,12 +101,14 @@ class DerivaServer (DerivaBinding):
         """
         return ErmrestAlias.connect(self, id)
 
-    def create_ermrest_alias(self, id=None, owner=None, alias_target=None):
+    def create_ermrest_alias(self, id=None, owner=None, alias_target=None, name=None, description=None):
         """Create an ERMrest catalog alias.
 
         :param id: The (str) id desired by the client (default None)
         :param owner: The initial (list of str) ACL desired by the client (default None)
         :param alias_target: The initial target catalog id binding desired by the client (default None)
+        :param name: Initial (str) catalog name if not None
+        :param description: Initial (str) catalog description if not None
 
         The new alias id will be returned in the response, and used
         in future alias access. The use of the id parameter
@@ -117,8 +133,13 @@ class DerivaServer (DerivaBinding):
         influences which client(s) are allowed to retry creation with
         the same id.
 
+        The name and description parameters are passed through to the
+        alias creation service to initialize those respective metadata
+        fields of the new aliase's registry entry. See ERMrest
+        documentation for more detail.
+
         """
-        return ErmrestAlias.create(self, id, owner, alias_target)
+        return ErmrestAlias.create(self, id, owner, alias_target, name, description)
 
 class ErmrestCatalogMutationError(Exception):
     pass
@@ -128,6 +149,11 @@ _clone_state_url = "tag:isrd.isi.edu,2018:clone-status"
 
 DEFAULT_PAGE_SIZE = 100000
 
+class ResolveRidResult (NamedTuple):
+    datapath: datapath.DataPath
+    table: ermrest_model.Table
+    rid: str
+
 
 class ErmrestCatalog(DerivaBinding):
     """Persistent handle for an ERMrest catalog.
@@ -198,15 +224,22 @@ class ErmrestCatalog(DerivaBinding):
         )
 
     @classmethod
-    def _digest_catalog_args(cls, id, owner):
+    def _digest_catalog_args(cls, id, owner, name=None, description=None, is_persistent=None, clone_source=None):
         rep = dict()
 
-        if isinstance(id, str):
-            rep['id'] = id
-        elif isinstance(id, (type(nochange), type(None))):
-            pass
-        else:
-            raise TypeError('id must be of type str or None or nochange, not %s' % type(id))
+        for v, k, typ in [
+                (id, 'id', str),
+                (name, 'name', str),
+                (description, 'description', str),
+                (is_persistent, 'is_persistent', bool),
+                (clone_source, 'clone_source', str),
+        ]:
+            if isinstance(v, typ):
+                rep[k] = v
+            elif isinstance(v, (type(nochange), type(None))):
+                pass
+            else:
+                raise TypeError('%s must be of type %s or None or nochange, not %s' % (k, typ.__name__, type(v)))
 
         if isinstance(owner, list):
             for e in owner:
@@ -221,12 +254,16 @@ class ErmrestCatalog(DerivaBinding):
         return rep
 
     @classmethod
-    def create(cls, deriva_server, id=None, owner=None):
+    def create(cls, deriva_server, id=None, owner=None, name=None, description=None, is_persistent=None, clone_source=None):
         """Create an ERMrest catalog and return the ERMrest catalog binding.
 
         :param deriva_server: The DerivaServer binding which hosts ermrest.
         :param id: The (str) id desired by the client (default None)
         :param owner: The initial (list of str) ACL desired by the client (default None)
+        :param name: Initial (str) catalog name if not None
+        :param description: Initial (str) catalog description if not None
+        :param is_persistent: Initial (bool) catalog persistence flag if not None
+        :param clone_source: Initial catalog clone_source if not None
 
         The new catalog id will be returned in the response, and used
         in future catalog access. The use of the id parameter
@@ -246,9 +283,18 @@ class ErmrestCatalog(DerivaBinding):
         influences which client(s) are allowed to retry creation with
         the same id.
 
+        The name, description, is_persistent, and clone_source
+        parameters are passed through to the catalog creation service
+        to initialize those respective metadata fields of the new
+        catalog's registry entry. See ERMrest documentation for more
+        detail. Authorization failures may occur when attempting to
+        set the is_persistent flag. By default, these fields are not
+        initialized in the catalog creation request, and they instead
+        receive server-assigned defaults.
+
         """
         path = '/ermrest/catalog'
-        r = deriva_server.post(path, json=cls._digest_catalog_args(id, owner))
+        r = deriva_server.post(path, json=cls._digest_catalog_args(id, owner, name, description, is_persistent, clone_source))
         r.raise_for_status()
         return cls.connect(deriva_server, r.json()['id'])
 
@@ -386,15 +432,75 @@ class ErmrestCatalog(DerivaBinding):
             return None
         return entity[0], entity[1]
 
+    def resolve_rid(self, rid: str, model: ermrest_model.Model=None, builder: datapath._CatalogWrapper=None) -> ResolveRidResult:
+        """Resolve a RID value to return a ResolveRidResult (a named tuple).
+
+        :param rid: The RID (str) to resolve
+        :param model: A result from self.getCatalogModel() to reuse
+        :param builder: A result from self.getPathBuilder() to reuse
+
+        Raises KeyError if RID is not found in the catalog.
+
+        The elements of the ResolveRidResult namedtuple provide more
+        information about the entity identified by the supplied RID in
+        this catalog:
+
+        - datapath: datapath instance for querying the resolved entity
+        - table: ermrest_model.Table instance containing the entity
+        - rid: normalized version of the input RID value
+
+        Example to simply retrieve entity content:
+
+           path, _, _ = catalog.resolve_rid('1-0000')
+           data = path.entities().fetch()[0]
+
+        """
+        if model is None:
+            model = self.getCatalogModel()
+        if builder is None:
+            builder = self.getPathBuilder()
+        try:
+            r = self.get('/entity_rid/%s' % urlquote(rid))
+            info = r.json()
+            sname = info['schema_name']
+            tname = info['table_name']
+            rid = info['RID']
+
+            ptable = builder.schemas[sname].tables[tname]
+
+            return ResolveRidResult(
+                ptable.path.filter(ptable.RID == rid),
+                model.schemas[sname].tables[tname],
+                rid
+            )
+        except requests.exceptions.HTTPError as e:
+            if e.response.status_code == requests.codes.not_found:
+                raise KeyError(rid)
+            raise
+
     def getAsFile(self,
                   path,
                   destfilename,
-                  headers=DEFAULT_HEADERS,
-                  callback=None,
-                  delete_if_empty=False,
-                  paged=False,
-                  page_size=DEFAULT_PAGE_SIZE,
-                  page_sort_columns=frozenset(["RID"])):
+                  headers = DEFAULT_HEADERS,
+                  callback = None,
+                  delete_if_empty = False,
+                  paged = False,
+                  page_size = DEFAULT_PAGE_SIZE,
+                  page_sort_columns = frozenset(["RID"])):
+        """
+           Deprecated, call `get_as_file` instead.
+        """
+        self.get_as_file(path, destfilename, headers, callback, delete_if_empty, paged, page_size, page_sort_columns)
+
+    def get_as_file(self,
+                    path,
+                    destfilename,
+                    headers=DEFAULT_HEADERS,
+                    callback=None,
+                    delete_if_empty=False,
+                    paged=False,
+                    page_size=DEFAULT_PAGE_SIZE,
+                    page_sort_columns=frozenset(["RID"])):
         """
            Retrieve catalog data streamed to destination file.
            Caller is responsible to clean up file even on error, when the file may or may not exist.
@@ -540,7 +646,7 @@ class ErmrestCatalog(DerivaBinding):
                     if buf == b"[]\n" or buf == b"{}\n":
                         delete_file = True
                 elif content_type == "text/csv":
-                    reader = csv.reader(codecs.iterdecode(destfile, 'utf-8') if not IS_PY2 else destfile)
+                    reader = csv.reader(codecs.iterdecode(destfile, 'utf-8'))
                     rowcount = 0
                     for row in reader:
                         rowcount += 1
@@ -603,7 +709,8 @@ class ErmrestCatalog(DerivaBinding):
                       copy_annotations=True,
                       copy_policy=True,
                       truncate_after=True,
-                      exclude_schemas=None):
+                      exclude_schemas=None,
+                      dst_properties=None):
         """Clone this catalog's content into dest_catalog, creating a new catalog if needed.
 
         :param dst_catalog: Destination catalog or None to request creation of new destination (default).
@@ -612,18 +719,33 @@ class ErmrestCatalog(DerivaBinding):
         :param copy_policy: Copy access-control policies when True (default).
         :param truncate_after: Truncate destination history after cloning when True (default).
         :param exclude_schemas: A list of schema names to exclude from the cloning process.
+        :param dst_properties: A dictionary of custom catalog-creation properties.
 
-        When dest_catalog is provided, attempt an idempotent clone,
+        When dst_catalog is provided, attempt an idempotent clone,
         assuming content MAY be partially cloned already using the
         same parameters. This routine uses a table-level annotation
         "tag:isrd.isi.edu,2018:clone-state" to save progress markers
         which help it restart efficiently if interrupted.
 
-        Cloning preserves source row RID values so that any RID-based
-        foreign keys are still valid. It is not generally advisable to
-        try to merge more than one source into the same clone, nor to
-        clone on top of rows generated locally in the destination,
-        since this could cause duplicate RID conflicts.
+        When dst_catalog is not provided, a new catalog is
+        provisioned. The optional dst_properties can customize
+        metadata properties during this step:
+
+        - name: str
+        - description: str (markdown-formatted)
+        - is_persistent: boolean
+
+        Cloning preserves source row RID values for application tables
+        so that any RID-based foreign keys are still valid. It is not
+        generally advisable to try to merge more than one source into
+        the same clone, nor to clone on top of rows generated locally
+        in the destination, since this could cause duplicate RID
+        conflicts.
+
+        Cloning does not preserve all RID values for special ERMrest
+        tables in the public schema (e.g. ERMrest_Client,
+        ERMrest_Group) but normal applications should only consider
+        the ID key of these tables.
 
         Truncation after cloning avoids retaining incremental
         snapshots which contain partial clones.
@@ -634,10 +756,33 @@ class ErmrestCatalog(DerivaBinding):
         session_config["allow_retry_on_all_methods"] = True
 
         if dst_catalog is None:
-            # TODO: refactor with DerivaServer someday
-            server = DerivaBinding(self._scheme, self._server, self._credentials, self._caching, session_config)
-            dst_id = server.post("/ermrest/catalog").json()["id"]
-            dst_catalog = ErmrestCatalog(self._scheme, self._server, dst_id, self._credentials, self._caching, session_config)
+            if dst_properties is not None:
+                if not isinstance(dst_properties, dict):
+                    raise TypeError('dst_properties must be of type dict or None, not %s' % (type(dst_properties),))
+            else:
+                dst_properties = {}
+            kwargs = {
+                "name": dst_properties.get('name', 'Clone of %r' % (self._catalog_id,)),
+                "description": dst_properties.get(
+                    'description',
+                    '''A cloned copy of catalog %r made with ErmrestCatalog.clone_catalog() using the following parameters:
+- `copy_data`: %r
+- `copy_annotations`: %r
+- `copy_policy`: %r
+- `truncate_after`: %r
+- `exclude_schemas`: %r
+''' % (
+    self._catalog_id,
+    copy_data,
+    copy_annotations,
+    copy_policy,
+    truncate_after,
+    exclude_schemas,
+)),
+                "clone_source": dst_properties.get('clone_source', self._catalog_id),
+            }
+            server = self.deriva_server
+            dst_catalog = server.create_ermrest_catalog(**kwargs)
 
         # set top-level config right away and find fatal usage errors...
         if copy_policy:
@@ -658,33 +803,29 @@ class ErmrestCatalog(DerivaBinding):
         fkeys_deferred = {}
         exclude_schemas = [] if exclude_schemas is None else exclude_schemas
 
-        def prune_parts(d):
-            if not copy_annotations and 'annotations' in d:
-                del d['annotations']
+        def prune_parts(d, *extra_victims):
+            victims = set(extra_victims)
+            # we will apply config as a second pass after extending dest model
+            # but loading bulk first may speed that up
+            if not copy_annotations:
+                victims |= {'annotations',}
             if not copy_policy:
-                if 'acls' in d:
-                    del d['acls']
-                if 'acl_bindings' in d:
-                    del d['acl_bindings']
+                victims |= {'acls', 'acl_bindings'}
+            for k in victims:
+                d.pop(k, None)
             return d
 
         def copy_sdef(s):
             """Copy schema definition structure with conditional parts for cloning."""
-            d = prune_parts(s.prejson())
-            if 'tables' in d:
-                del d['tables']
+            d = prune_parts(s.prejson(), 'tables')
             return d
 
         def copy_tdef_core(t):
             """Copy table definition structure with conditional parts excluding fkeys."""
-            d = prune_parts(t.prejson())
+            d = prune_parts(t.prejson(), 'foreign_keys')
             d['column_definitions'] = [ prune_parts(c) for c in d['column_definitions'] ]
             d['keys'] = [ prune_parts(c) for c in d.get('keys', []) ]
-            if 'foreign_keys' in d:
-                del d['foreign_keys']
-            if 'annotations' not in d:
-                d['annotations'] = {}
-                d['annotations'][_clone_state_url] = 1 if copy_data else None
+            d.setdefault('annotations', {})[_clone_state_url] = 1 if copy_data else None
             return d
 
         def copy_tdef_fkeys(t):
@@ -692,7 +833,7 @@ class ErmrestCatalog(DerivaBinding):
             def check(fkdef):
                 for fkc in fkdef['referenced_columns']:
                     if fkc['schema_name'] == 'public' \
-                       and fkc['table_name'] in {'ERMrest_Client', 'ERMrest_Group'} \
+                       and fkc['table_name'] in {'ERMrest_Client', 'ERMrest_Group', 'ERMrest_RID_Lease'} \
                        and fkc['column_name'] == 'RID':
                         raise ValueError("Cannot clone catalog with foreign key reference to %(schema_name)s:%(table_name)s:%(column_name)s" % fkc)
                 return fkdef
@@ -739,6 +880,12 @@ class ErmrestCatalog(DerivaBinding):
                     clone_states[(sname, tname)] = 1 if copy_data else None
                     fkeys_deferred[(sname, tname)] = copy_tdef_fkeys(table)
                 else:
+                    if dst_model.schemas[sname].tables[tname].foreign_keys:
+                        # assume that presence of any destination foreign keys means we already loaded deferred_fkeys
+                        copy_data = False
+                    else:
+                        fkeys_deferred[(sname, tname)] = copy_tdef_fkeys(table)
+
                     src_columns = { c.name: c for c in table.column_definitions }
                     dst_columns = { c.name: c for c in dst_model.schemas[sname].tables[tname].column_definitions }
 
@@ -764,11 +911,8 @@ class ErmrestCatalog(DerivaBinding):
                             raise ValueError("Destination key %s.%s(%s) does not exist in source catalog." % (sname, tname, ', '.join(utuple)))
 
                     clone_states[(sname, tname)] = dst_model.schemas[sname].tables[tname].annotations.get(_clone_state_url)
-                    if dst_model.schemas[sname].tables[tname].foreign_keys:
-                        # assume that presence of any destination foreign keys means we already completed
-                        return dst_catalog
-                    else:
-                        fkeys_deferred[(sname, tname)] = copy_tdef_fkeys(table)
+
+        clone_states[('public', 'ERMrest_RID_Lease')] = None # never try to sync leases
 
         # apply the stage 1 model to the destination in bulk
         if new_model:
@@ -822,38 +966,8 @@ class ErmrestCatalog(DerivaBinding):
                 }:
                     # special sync behavior for magic ermrest tables
                     # HACK: these are assumed small enough to join via local merge of arrays
-                    want = sorted(self.get("/entity/%s?limit=none" % tname_uri).json(), key=lambda r: r['ID'])
-                    have = sorted(dst_catalog.get("/entity/%s?limit=none" % tname_uri).json(), key=lambda r: r['ID'])
-                    create = []
-                    update = []
-
-                    pos_want = 0
-                    pos_have = 0
-                    while pos_want < len(want):
-                        while pos_have < len(have) and have[pos_have]['ID'] < want[pos_want]['ID']:
-                            # dst-only rows will be retained as is
-                            pos_have += 1
-                        if pos_have >= len(have) or have[pos_have]['ID'] > want[pos_want]['ID']:
-                            # src-only rows will be inserted
-                            create.append(want[pos_want])
-                            pos_want += 1
-                        else:
-                            # overlapping rows will be updated
-                            update.append(want[pos_want])
-                            pos_want += 1
-
-                    dst_catalog.post("/entity/%s?nondefaults=RCT,RCB" % tname_uri, json=create)
-                    dst_catalog.put(
-                        "/attributegroup/%s/ID;%s" % (
-                            tname_uri,
-                            ",".join([
-                                urlquote(c.name)
-                                for c in src_model.schemas[sname].tables[tname].column_definitions
-                                if c.name not in {'RID', 'RMT', 'RMB', 'ID'}
-                            ])
-                        ),
-                        json=update
-                    )
+                    page = self.get("/entity/%s?limit=none" % tname_uri).json()
+                    dst_catalog.post("/entity/%s?onconflict=skip" % tname_uri, json=page)
 
                     # record our progress on catalog in case we fail part way through
                     dst_catalog.put(
@@ -873,6 +987,84 @@ class ErmrestCatalog(DerivaBinding):
         if new_fkeys:
             dst_catalog.post("/schema", json=new_fkeys)
 
+        # copy over configuration in stage 3
+        # we need to do this after deferred_fkeys to handle acl_bindings projections with joins
+        dst_model = dst_catalog.getCatalogModel()
+
+        for sname, src_schema in src_model.schemas.items():
+            if sname in exclude_schemas:
+                continue
+            dst_schema = dst_model.schemas[sname]
+
+            if copy_annotations:
+                dst_schema.annotations.clear()
+                dst_schema.annotations.update(src_schema.annotations)
+
+            if copy_policy:
+                dst_schema.acls.clear()
+                dst_schema.acls.update(src_schema.acls)
+
+            for tname, src_table in src_schema.tables.items():
+                dst_table = dst_schema.tables[tname]
+
+                if copy_annotations:
+                    merged = dict(src_table.annotations)
+                    if _clone_state_url in dst_table.annotations:
+                        merged[_clone_state_url] = dst_table.annotations[_clone_state_url]
+                    dst_table.annotations.clear()
+                    dst_table.annotations.update(merged)
+
+                if copy_policy:
+                    dst_table.acls.clear()
+                    dst_table.acls.update(src_table.acls)
+                    dst_table.acl_bindings.clear()
+                    dst_table.acl_bindings.update(src_table.acl_bindings)
+
+                for cname, src_col in src_table.columns.elements.items():
+                    dst_col = dst_table.columns[cname]
+
+                    if copy_annotations:
+                        dst_col.annotations.clear()
+                        dst_col.annotations.update(src_col.annotations)
+
+                    if copy_policy:
+                        dst_col.acls.clear()
+                        dst_col.acls.update(src_col.acls)
+                        dst_col.acl_bindings.clear()
+                        dst_col.acl_bindings.update(src_col.acl_bindings)
+
+                for src_key in src_table.keys:
+                    dst_key = dst_table.key_by_columns([ col.name for col in src_key.unique_columns ])
+
+                    if copy_annotations:
+                        dst_key.annotations.clear()
+                        dst_key.annotations.update(src_key.annotations)
+
+                def xlate_column_map(fkey):
+                    dst_from_table = dst_table
+                    dst_to_schema = dst_model.schemas[fkey.pk_table.schema.name]
+                    dst_to_table = dst_to_schema.tables[fkey.pk_table.name]
+                    return {
+                        dst_from_table._own_column(from_col.name): dst_to_table._own_column(to_col.name)
+                        for from_col, to_col in fkey.column_map.items()
+                    }
+
+                for src_fkey in src_table.foreign_keys:
+                    dst_fkey = dst_table.fkey_by_column_map(xlate_column_map(src_fkey))
+
+                    if copy_annotations:
+                        dst_fkey.annotations.clear()
+                        dst_fkey.annotations.update(src_fkey.annotations)
+
+                    if copy_policy:
+                        dst_fkey.acls.clear()
+                        dst_fkey.acls.update(src_fkey.acls)
+                        dst_fkey.acl_bindings.clear()
+                        dst_fkey.acl_bindings.update(src_fkey.acl_bindings)
+
+        # send all the config changes to the server
+        dst_model.apply()
+
         # truncate cloning history
         if truncate_after:
             snaptime = dst_catalog.get("/").json()["snaptime"]
@@ -946,8 +1138,8 @@ class ErmrestAlias(DerivaBinding):
         )
 
     @classmethod
-    def _digest_alias_args(cls, id, owner, alias_target):
-        rep = ErmrestCatalog._digest_catalog_args(id, owner)
+    def _digest_alias_args(cls, id, owner, alias_target, name, description):
+        rep = ErmrestCatalog._digest_catalog_args(id, owner, name, description)
 
         if isinstance(alias_target, (str, type(None))):
             rep['alias_target'] = alias_target
@@ -959,13 +1151,15 @@ class ErmrestAlias(DerivaBinding):
         return rep
 
     @classmethod
-    def create(cls, deriva_server, id=None, owner=None, alias_target=None):
+    def create(cls, deriva_server, id=None, owner=None, alias_target=None, name=None, description=None):
         """Create an ERMrest catalog alias.
 
         :param deriva_server: The DerivaServer binding which hosts ermrest
         :param id: The (str) id desired by the client (default None)
         :param owner: The initial (list of str) ACL desired by the client (default None)
         :param alias_target: The initial target catalog id desired by the client (default None)
+        :param name: Initial (str) catalog name if not None
+        :param description: Initial (str) catalog description if not None
 
         The new alias id will be returned in the response, and used
         in future alias access. The use of the id parameter
@@ -990,9 +1184,14 @@ class ErmrestAlias(DerivaBinding):
         influences which client(s) are allowed to retry creation with
         the same id.
 
+        The name and description parameters are passed through to the
+        alias creation service to initialize those respective metadata
+        fields of the new aliase's registry entry. See ERMrest
+        documentation for more detail.
+
         """
         path = '/ermrest/alias'
-        r = deriva_server.post(path, json=cls._digest_alias_args(id, owner, alias_target))
+        r = deriva_server.post(path, json=cls._digest_alias_args(id, owner, alias_target, name, description))
         r.raise_for_status()
         return cls.connect(deriva_server, r.json()['id'])
 
