@@ -13,29 +13,10 @@ from requests.adapters import HTTPAdapter
 from urllib3.util.retry import Retry
 from collections import OrderedDict
 from distutils import util as du_util
+from urllib.parse import quote as _urlquote, unquote as urlunquote
+from urllib.parse import urlparse, urlsplit, urlunsplit, urljoin
+from http.cookiejar import MozillaCookieJar
 
-IS_PY2 = (sys.version_info[0] == 2)
-IS_PY3 = (sys.version_info[0] == 3)
-
-
-def force_unicode_py2(s):
-    """Reliably return a Unicode string given a possible unicode or byte string"""
-    if isinstance(s, str):
-        return s.decode("utf-8")
-    else:
-        return unicode(s)
-
-
-if IS_PY3:
-    from urllib.parse import quote as _urlquote, unquote as urlunquote
-    from urllib.parse import urlparse, urlsplit, urlunsplit, urljoin
-    from http.cookiejar import MozillaCookieJar
-    force_unicode = str
-else:
-    from urllib import quote as _urlquote, unquote as urlunquote
-    from urlparse import urlparse, urlsplit, urlunsplit, urljoin
-    from cookielib import MozillaCookieJar
-    force_unicode = force_unicode_py2
 
 Kilobyte = 1024
 Megabyte = Kilobyte ** 2
@@ -57,7 +38,7 @@ DEFAULT_CREDENTIAL_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'credential.json')
 DEFAULT_GLOBUS_CREDENTIAL_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'globus-credential.json')
 DEFAULT_CONFIG_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'config.json')
 DEFAULT_COOKIE_JAR_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'cookies.txt')
-DEFAULT_REQUESTS_TIMEOUT = (6.3, 63)  # (connect, read), in seconds
+DEFAULT_REQUESTS_TIMEOUT = (6, 63)  # (connect, read), integer in seconds
 DEFAULT_SESSION_CONFIG = {
     "timeout": DEFAULT_REQUESTS_TIMEOUT,
     "retry_connect": 2,
@@ -67,7 +48,8 @@ DEFAULT_SESSION_CONFIG = {
     "allow_retry_on_all_methods": False,
     "cookie_jar": DEFAULT_COOKIE_JAR_FILE,
     "max_request_size": DEFAULT_MAX_REQUEST_SIZE,
-    "max_chunk_limit": DEFAULT_MAX_CHUNK_LIMIT
+    "max_chunk_limit": DEFAULT_MAX_CHUNK_LIMIT,
+    "bypass_cert_verify_host_list": []
 }
 OAUTH2_SCOPES_KEY = "oauth2_scopes"
 DEFAULT_CONFIG = {
@@ -203,16 +185,20 @@ def get_new_requests_session(url=None, session_config=DEFAULT_SESSION_CONFIG):
                     read=session_config['retry_read'],
                     backoff_factor=session_config['retry_backoff_factor'],
                     status_forcelist=session_config['retry_status_forcelist'],
-                    method_whitelist=Retry.DEFAULT_METHOD_WHITELIST if
+                    allowed_methods=Retry.DEFAULT_ALLOWED_METHODS if
                     # Passing False to method_whitelist means allow all methods
                     not session_config.get("allow_retry_on_all_methods", False) else False,
                     raise_on_status=True)
     adapter = TimeoutHTTPAdapter(timeout=session_config.get("timeout", DEFAULT_REQUESTS_TIMEOUT), max_retries=retries)
+    session.mount('http://', adapter)
+    session.mount('https://', adapter)
+
     if url:
-        session.mount(url, adapter)
-    else:
-        session.mount('http://', adapter)
-        session.mount('https://', adapter)
+        # allow whitelisted hosts to bypass SSL cert verification
+        upr = urlparse(url)
+        bypass_cert_verify_host_list = session_config.get("bypass_cert_verify_host_list", [])
+        if upr.scheme == "https" and upr.hostname in bypass_cert_verify_host_list:
+            session.verify = False
 
     return session
 
@@ -237,8 +223,6 @@ def write_config(config_file=DEFAULT_CONFIG_FILE, config=DEFAULT_CONFIG):
     make_dirs(config_dir, mode=0o750)
     with io.open(config_file, 'w', newline='\n', encoding='utf-8') as cf:
         config_data = json.dumps(config, ensure_ascii=False, indent=2)
-        if IS_PY2 and isinstance(config_data, str):
-            config_data = unicode(config_data, 'utf-8')
         cf.write(config_data)
         cf.close()
 
@@ -275,8 +259,6 @@ def write_credential(credential_file=DEFAULT_CREDENTIAL_FILE, credential=DEFAULT
     with lock_file(credential_file, mode='w', exclusive=True) as cf:
         os.chmod(credential_file, 0o600)
         credential_data = json.dumps(credential, ensure_ascii=False, indent=2)
-        if IS_PY2 and isinstance(credential_data, str):
-            credential_data = unicode(credential_data, 'utf-8')
         cf.write(credential_data)
         cf.flush()
         os.fsync(cf.fileno())
@@ -324,7 +306,7 @@ def get_oauth_scopes_for_host(host,
                 result = scopes
                 break
     if not result or force_refresh:
-        session = get_new_requests_session(session_config=DEFAULT_SESSION_CONFIG)
+        session = get_new_requests_session(url, session_config=config.get("session", DEFAULT_SESSION_CONFIG))
         try:
             r = session.get(url, headers=DEFAULT_HEADERS)
             r.raise_for_status()
@@ -512,11 +494,16 @@ tag = AttrDict({
     'visible_foreign_keys': 'tag:isrd.isi.edu,2016:visible-foreign-keys',
     'export':             'tag:isrd.isi.edu,2016:export',
     'export_2019':        'tag:isrd.isi.edu,2019:export',
+    'export_fragment_definitions': 'tag:isrd.isi.edu,2021:export-fragment-definitions',    
     'asset':              'tag:isrd.isi.edu,2017:asset',
     'citation':           'tag:isrd.isi.edu,2018:citation',
     'required':           'tag:isrd.isi.edu,2018:required',
     'indexing_preferences': 'tag:isrd.isi.edu,2018:indexing-preferences',
     'bulk_upload':        'tag:isrd.isi.edu,2017:bulk-upload',
     'chaise_config':      'tag:isrd.isi.edu,2019:chaise-config',
-    'source_definitions': 'tag:isrd.isi.edu,2019:source-definitions'
+    'source_definitions': 'tag:isrd.isi.edu,2019:source-definitions',
+    'indexing_preferences': 'tag:isrd.isi.edu,2018:indexing-preferences',
+    'google_dataset':     'tag:isrd.isi.edu,2021:google-dataset',
+    'column_defaults':    'tag:isrd.isi.edu,2023:column-defaults',
+    'viz_3d_display':     'tag:isrd.isi.edu,2021:viz-3d-display',
 })
