@@ -5,16 +5,18 @@ import json
 import logging
 import platform
 import traceback
-import importlib
+import time
 import datetime
 import tzlocal
+import webbrowser
 from pprint import pprint
 from requests.exceptions import HTTPError, ConnectionError
 from bdbag.fetch.auth import keychain as bdbkc
 from deriva.core import __version__ as VERSION, DEFAULT_CONFIG_PATH, DEFAULT_GLOBUS_CREDENTIAL_FILE, urlparse, urljoin,\
     read_config, format_exception, BaseCLI, get_oauth_scopes_for_host, get_new_requests_session
 from deriva.core.utils import eprint
-from globus_sdk import ConfidentialAppAuthClient, AuthClient, AccessTokenAuthorizer, GlobusError, GlobusAPIError
+from globus_sdk import ConfidentialAppAuthClient, AuthClient, AccessTokenAuthorizer, RefreshTokenAuthorizer, \
+    GlobusError, GlobusAPIError, AuthAPIError
 from fair_research_login.client import NativeClient, LoadError, NoSavedTokens, TokensExpired
 
 NATIVE_APP_CLIENT_ID = "8ef15ba9-2b4a-469c-a163-7fd910c9d111"
@@ -24,7 +26,7 @@ GROUPS_SCOPE_NAME = "urn:globus:auth:scope:groups.api.globus.org:view_my_groups_
 PUBLIC_GROUPS_API_URL = 'https://groups.api.globus.org/v2/groups/my_groups'
 CLIENT_CRED_FILE = '/home/secrets/oauth2/client_secret_globus.json'
 DEFAULT_SCOPES = ["openid", GROUPS_SCOPE_NAME]
-
+LOGOUT_URL = "https://app.globus.org/logout"
 
 class UsageException(ValueError):
     """Usage exception.
@@ -51,7 +53,7 @@ class GlobusAuthUtil:
         client_id = kwargs.get("client_id")
         client_secret = kwargs.get("client_secret")
         if not (client_id and client_secret):
-            cred_file = kwargs.get("credential_file", CLIENT_CRED_FILE)
+            cred_file = kwargs.get("credential_file", CLIENT_CRED_FILE) or CLIENT_CRED_FILE
             if os.path.isfile(cred_file):
                 creds = read_config(cred_file)
                 if creds:
@@ -59,6 +61,8 @@ class GlobusAuthUtil:
                     if client:
                         client_id = client.get('client_id')
                         client_secret = client.get('client_secret')
+            else:
+                logging.warning("No Globus client credential file found at: %s" % cred_file)
 
         if not (client_id and client_secret):
             logging.warning("Client ID and secret not specified and/or could not be determined.")
@@ -92,27 +96,27 @@ class GlobusAuthUtil:
             raise UsageException("A supported scope argument is required.")
 
         r = self.client.post("/v2/api/clients/{client_id}/scopes".format(client_id=self.client_id),
-                             json_body=self.from_json(scope))
+                             data=self.from_json(scope))
         return r.data
 
     def update_scope(self, scope_id, scope):
         if not (scope_id and scope):
             raise UsageException("The scope_id and scope arguments are required.")
 
-        r = self.client.put("/v2/api/scopes/{scope_id}".format(scope_id=scope_id), json_body=self.from_json(scope))
+        r = self.client.put("/v2/api/scopes/{scope_id}".format(scope_id=scope_id), data=self.from_json(scope))
         return r.data
 
     def add_fqdn_to_client(self, fqdn):
         if not fqdn:
             raise UsageException("A fqdn (fully qualified domain name) argument is required.")
         r = self.client.post('/v2/api/clients/{client_id}/fqdns'.format(client_id=self.client_id),
-                             json_body={'fqdn': fqdn})
+                             data={'fqdn': fqdn})
         return r.data
 
     def get_client_for_fqdn(self, fqdn):
         if not fqdn:
             raise UsageException("A fqdn (fully qualified domain name) argument is required.")
-        r = self.client.get('/v2/api/clients?fqdn={fqdn}'.format(fqdn=fqdn))
+        r = self.client.get('/v2/api/clients', query_params={"fqdn": fqdn})
         return r.data
 
     def get_client(self, client_id):
@@ -139,12 +143,11 @@ class GlobusAuthUtil:
         if not token:
             raise UsageException("A token argument is required.")
         additional_params = {"access_type": "offline"} if refresh else None
-        r = self.client.oauth2_get_dependent_tokens(token, additional_params)
+        r = self.client.oauth2_get_dependent_tokens(token, additional_params=additional_params)
         return r.data
 
     def new_client(self, client):
-        r = self.client.post("/v2/api/clients",
-                             json_body=self.from_json(client))
+        r = self.client.post("/v2/api/clients", data=self.from_json(client))
         return r.data
 
     def create_client(self,
@@ -183,7 +186,7 @@ class GlobusAuthUtil:
             raise UsageException("A client argument is required.")
 
         r = self.client.put("/v2/api/clients/{client_id}".format(client_id=client_id if client_id else self.client_id),
-                            json_body=self.from_json(client))
+                            data=self.from_json(client))
         return r.data
 
     def delete_client(self, client_id):
@@ -210,7 +213,7 @@ class GlobusAuthUtil:
     def get_scopes_by_name(self, scope_name):
         if not scope_name:
             raise UsageException("A scope_name argument is required.")
-        scopes = self.client.get('/v2/api/scopes?scope_strings={sname}'.format(sname=scope_name))
+        scopes = self.client.get('/v2/api/scopes', query_params={"scope_strings": {scope_name}})
         if scopes is None:
             return None
         else:
@@ -219,7 +222,7 @@ class GlobusAuthUtil:
     def get_scopes_by_id(self, id_string):
         if not id_string:
             raise UsageException("A id_string argument is required.")
-        scopes = self.client.get('/v2/api/scopes?ids={ids}'.format(ids=id_string))
+        scopes = self.client.get('/v2/api/scopes', query_params={"ids": id_string})
         if scopes is None:
             return None
         else:
@@ -270,7 +273,7 @@ class GlobusAuthUtil:
             }
         }
 
-        r = self.client.put('/v2/api/clients/{client_id}'.format(client_id=self.client_id), json_body=d)
+        r = self.client.put('/v2/api/clients/{client_id}'.format(client_id=self.client_id), data=d)
         return r.data
 
     def add_dependent_scopes(self, parent_scope_name, child_scopes,
@@ -309,8 +312,7 @@ class GlobusAuthUtil:
             }
         }
         print(str(d))
-        r = self.client.put('/v2/api/scopes/{i}'.format(i=parent_scope_id),
-                            json_body=d)
+        r = self.client.put('/v2/api/scopes/{i}'.format(i=parent_scope_id), data=d)
         return r.data
 
     def create_scope_with_deps(self, name, description, suffix, dependent_scopes=[], advertised=True,
@@ -340,7 +342,7 @@ class GlobusAuthUtil:
             }
         }
 
-        r = self.client.post("/v2/api/clients/{client_id}/scopes".format(client_id=self.client_id), json_body=scope)
+        r = self.client.post("/v2/api/clients/{client_id}/scopes".format(client_id=self.client_id), data=scope)
         return r.data
 
     def delete_scope(self, scope_name):
@@ -585,12 +587,47 @@ class GlobusNativeLogin:
         if not scopes:
             return None
         try:
-            return self.client.load_tokens(scopes)
+            tokens = self.client.load_tokens(list(scopes))
+            return self.refresh_tokens_preemptively(tokens, time_before_expiry=86400)
         except LoadError as e:
             logging.debug("Unable to load or find tokens for specified scopes [%s]: %s" %
                           (scopes, format_exception(e)))
             return None
 
+    def refresh_tokens_preemptively(self, tokens, time_before_expiry=0):
+        needs_refresh = {}
+
+        for rs, ts in {t: ts for t, ts in tokens.items() if bool(ts['refresh_token'])}.items():
+            expiry_threshold = int(time.time()) + time_before_expiry
+            if ts['expires_at_seconds'] < expiry_threshold:
+                needs_refresh.update({rs:ts})
+
+        if not needs_refresh:
+            return tokens
+
+        for rs, ts in needs_refresh.items():
+            authorizer = RefreshTokenAuthorizer(
+                ts['refresh_token'],
+                self.client.client,
+                access_token=ts['access_token'],
+                expires_at=int(time.time()),
+                # setting the expires_at to the current time is vicious hack but necessary since the globus_sdk
+                # client doesn't allow you to refresh an access token unless it has already expired
+            )
+            try:
+                authorizer.ensure_valid_token()
+                ts['access_token'] = authorizer.access_token
+                ts['expires_at_seconds'] = authorizer.expires_at
+            except AuthAPIError as e:
+                if e.message == 'invalid_grant':
+                    logging.warning('Refresh Token expired for resource server: %s', rs)
+
+        self.client.save_tokens(needs_refresh)
+        result = {rs: ts for rs, ts in tokens.items() if rs not in needs_refresh}
+        result.update(needs_refresh)
+
+        return result
+
     def hosts_to_scope_map(self,
                            hosts,
                            match_scope_tag=None,
@@ -661,6 +698,7 @@ class GlobusNativeLogin:
             access_token = token.get("access_token")
             if (token_scope == scope) and (access_token is not None):
                 return access_token
+        return None
 
     def update_bdbag_keychain(self, token=None, host=None, keychain_file=None, allow_redirects=False, delete=False):
         if (token is None) or (host is None):
@@ -698,7 +736,7 @@ class GlobusNativeLogin:
             prefill_named_grant = self.client.app_name + " with requested scopes [%s] " % ", ".join(scopes)
         tokens = self.client.login(no_local_server=no_local_server,
                                    no_browser=no_browser,
-                                   requested_scopes=scopes,
+                                   requested_scopes=list(scopes),
                                    refresh_tokens=refresh_tokens,
                                    prefill_named_grant=prefill_named_grant,
                                    query_params=additional_params,
@@ -710,7 +748,8 @@ class GlobusNativeLogin:
                     self.update_bdbag_keychain(token=access_token, host=host, keychain_file=bdbag_keychain_file)
         return tokens
 
-    def logout(self, hosts=(), requested_scopes=(), exclude_defaults=False, bdbag_keychain_file=None):
+    def logout(self, hosts=(), requested_scopes=(), exclude_defaults=False, bdbag_keychain_file=None,
+               include_browser_logout=False):
         tokens = self.client._load_raw_tokens()
 
         scopes = set(requested_scopes)
@@ -738,6 +777,14 @@ class GlobusNativeLogin:
         self.client.revoke_token_set(token_set)
         self.client.token_storage.clear_tokens(scopes)
 
+        if include_browser_logout:
+            if bool(os.environ.get('SSH_TTY') or os.environ.get('SSH_CONNECTION')):
+                print("Browser logout flow not available with remote session. Visit %s to logout of Globus completely."
+                      % LOGOUT_URL)
+                return
+
+            webbrowser.open(LOGOUT_URL, new=1)
+
 
 class DerivaGlobusAuthUtilCLIException(Exception):
     def __init__(self, message):
@@ -1079,7 +1126,8 @@ class DerivaGlobusAuthUtilCLI(BaseCLI):
             self.gnl.logout(hosts=[args.host] if args.host else [],
                             requested_scopes=args.requested_scopes,
                             exclude_defaults=args.exclude_defaults,
-                            bdbag_keychain_file=args.bdbag_keychain_file)
+                            bdbag_keychain_file=args.bdbag_keychain_file,
+                            include_browser_logout=args.full)
             return "You have been logged out."
 
         parser = self.subparsers.add_parser("logout", help="Revoke and clear tokens. If no arguments are specified, "
@@ -1099,6 +1147,9 @@ class DerivaGlobusAuthUtilCLI(BaseCLI):
                                  "default set of scopes: [%s]" % ", ".join(DEFAULT_SCOPES))
         parser.add_argument('--bdbag-keychain-file', metavar='<file>',
                             help="Non-default path to a bdbag keychain file.")
+        parser.add_argument('--full', action="store_true",
+                            help="Also launch the default system web browser to logout of Globus and clear cached "
+                                 "identity-related browser cookies.")
         parser.set_defaults(func=logout)
 
     def user_info_init(self):
