@@ -1,8 +1,7 @@
 import os
 import errno
-import certifi
 import requests
-from deriva.core import urlsplit, get_new_requests_session, stob, make_dirs, DEFAULT_SESSION_CONFIG
+from deriva.core import urlsplit, get_new_requests_session, stob, make_dirs, format_exception, DEFAULT_SESSION_CONFIG
 from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError, \
     DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError
 from deriva.transfer.download.processors.base_processor import BaseProcessor, \
@@ -21,8 +20,8 @@ class BaseQueryProcessor(BaseProcessor):
         self.catalog = kwargs["catalog"]
         self.store = kwargs["store"]
         self.base_path = kwargs["base_path"]
-        self.query = self.parameters["query_path"]
-        if self.envars:
+        self.query = self.parameters.get("query_path", "")
+        if self.query and self.envars:
             self.query = self.query.format(**self.envars)
         self.sub_path = self.parameters.get("output_path")
         self.output_filename = self.parameters.get("output_filename")
@@ -59,6 +58,9 @@ class BaseQueryProcessor(BaseProcessor):
         return self.outputs
 
     def catalogQuery(self, headers=None, as_file=True):
+        if not self.query:
+            return {}
+
         if not headers:
             headers = self.HEADERS.copy()
         else:
@@ -69,23 +71,23 @@ class BaseQueryProcessor(BaseProcessor):
             make_dirs(output_dir)
         try:
             if as_file:
-                return self.catalog.getAsFile(self.query, self.output_abspath,
-                                              headers=headers,
-                                              callback=self.callback,
-                                              delete_if_empty=True,
-                                              paged=self.paged_query,
-                                              page_size=self.paged_query_size,
-                                              page_sort_columns=self.paged_query_sort_columns)
+                return self.catalog.get_as_file(self.query, self.output_abspath,
+                                                headers=headers,
+                                                callback=self.callback,
+                                                delete_if_empty=True,
+                                                paged=self.paged_query,
+                                                page_size=self.paged_query_size,
+                                                page_sort_columns=self.paged_query_sort_columns)
             else:
                 return self.catalog.get(self.query, headers=headers).json()
         except requests.HTTPError as e:
             if e.response.status_code == 401:
-                raise DerivaDownloadAuthenticationError(e)
+                raise DerivaDownloadAuthenticationError(format_exception(e))
             if e.response.status_code == 403:
-                raise DerivaDownloadAuthorizationError(e)
+                raise DerivaDownloadAuthorizationError(format_exception(e))
             if as_file:
                 os.remove(self.output_abspath)
-            raise DerivaDownloadError("Error executing catalog query: %s" % e)
+            raise DerivaDownloadError("Error executing catalog query: %s" % format_exception(e))
         except Exception:
             if as_file:
                 os.remove(self.output_abspath)
@@ -135,35 +137,36 @@ class BaseQueryProcessor(BaseProcessor):
             if not (urlparts.scheme and urlparts.netloc):
                 urlparts = urlsplit(self.catalog.get_server_uri())
                 server_uri = urlparts.scheme + "://" + urlparts.netloc
-                url = ''.join([server_uri, url])
+                sep = "/" if not url.startswith("/") else ""
+                url = ''.join([server_uri, sep, url])
 
         return url
 
-    def getExternalSession(self, host):
-        sessions = self.sessions
+    def getExternalSession(self, url):
         auth_params = self.kwargs.get("auth_params", dict())
         cookies = auth_params.get("cookies")
         auth_url = auth_params.get("auth_url")
         login_params = auth_params.get("login_params")
         session_config = self.kwargs.get("session_config")
 
-        session = sessions.get(host)
+        host = urlsplit(url).hostname
+        session = self.sessions.get(host)
         if session is not None:
             return session
 
         if not session_config:
             session_config = DEFAULT_SESSION_CONFIG
-        session = get_new_requests_session(session_config=session_config)
+        session = get_new_requests_session(url, session_config=session_config)
 
         if cookies:
             session.cookies.update(cookies)
         if login_params and auth_url:
-            r = session.post(auth_url, data=login_params, verify=certifi.where())
+            r = session.post(auth_url, data=login_params)
             if r.status_code > 203:
                 raise DerivaDownloadError(
                     'GetExternalSession Failed with Status Code: %s\n%s\n' % (r.status_code, r.text))
 
-        sessions[host] = session
+        self.sessions[host] = session
         return session
 
     def create_default_paths(self):
@@ -175,8 +178,9 @@ class BaseQueryProcessor(BaseProcessor):
                                                                      envars=self.envars)
 
     def __del__(self):
-        for session in self.sessions.values():
-            session.close()
+        if self.sessions:
+            for session in self.sessions.values():
+                session.close()
 
 
 class CSVQueryProcessor(BaseQueryProcessor):
@@ -206,10 +210,30 @@ class JSONStreamQueryProcessor(BaseQueryProcessor):
 class JSONEnvUpdateProcessor(BaseQueryProcessor):
     def __init__(self, envars=None, **kwargs):
         super(JSONEnvUpdateProcessor, self).__init__(envars, **kwargs)
+        self.query_keys = self.parameters.get("query_keys")
 
     def process(self):
         resp = self.catalogQuery(headers={'accept': "application/json"}, as_file=False)
         if resp:
-            self.envars.update(resp[0])
+            if isinstance(resp, list):
+                resp = resp[0]
+            if self.query_keys is not None:
+                results = {key: resp[key] for key in self.query_keys}
+            else:
+                results = resp
+            self.envars.update(results)
             self._urlencode_envars()
         return {}
+
+
+class CreateDirProcessor(JSONEnvUpdateProcessor):
+    def __init__(self, envars=None, **kwargs):
+        super(CreateDirProcessor, self).__init__(envars, **kwargs)
+        self.ext = ""
+
+    def process(self):
+        super(CreateDirProcessor, self).process()
+        self.create_default_paths()
+        make_dirs(self.output_abspath)
+
+        return self.outputs
