@@ -2,21 +2,24 @@ import io
 import os
 import re
 import sys
-import errno
+import datetime
 import json
 import shutil
 import tempfile
+import pathlib
 import logging
 import platform
-from collections import OrderedDict, namedtuple
+import signal
+from collections import OrderedDict
 from deriva.core import ErmrestCatalog, HatracStore, HatracJobAborted, HatracJobPaused, \
     HatracJobTimeout, urlquote, urlparse, stob, format_exception, get_credential, read_config, write_config, \
-    copy_config, resource_path, make_dirs, lock_file, DEFAULT_CHUNK_SIZE, IS_PY2, __version__ as VERSION
+    copy_config, resource_path, make_dirs, lock_file, DEFAULT_CHUNK_SIZE, __version__ as VERSION
 from deriva.core import DEFAULT_SESSION_CONFIG, DEFAULT_CREDENTIAL_FILE
 from deriva.core.utils import hash_utils as hu, mime_utils as mu, version_utils as vu
 from deriva.transfer.upload import *
 from deriva.transfer.upload.processors import find_processor
 from deriva.transfer.upload.processors.base_processor import *
+from deriva.transfer.upload.processors.archive_processor import ArchiveProcessor
 
 try:
     from os import scandir, walk
@@ -31,12 +34,13 @@ class Enum(tuple):
 
 
 UploadState = Enum(["Success", "Failed", "Pending", "Running", "Paused", "Aborted", "Cancelled", "Timeout"])
-FileUploadState = namedtuple("FileUploadState", ["State", "Status"])
-UploadMetadataReservedKeyNames = ["URI", "file_name", "file_ext", "file_size", "content-disposition", "md5", "sha256",
-                                  "md5_base64", "sha256_base64", "schema", "table", "target_table"]
+UploadMetadataReservedKeyNames = [
+    "URI", "file_name", "file_ext", "file_size", "base_path", "base_name", "content-disposition", "md5", "sha256",
+    "md5_base64", "sha256_base64", "schema", "table", "target_table", "_upload_year_", "_upload_month_", "_upload_day_",
+    "_upload_time_", "_identity_id", "_identity_display_name", "_identity_full_name", "_identity_email"]
 
 DefaultConfig = {
-  "version_compatibility": [">=%s" % VERSION],
+  "version_compatibility": [[">=%s" % VERSION]],
   "version_update_url": "https://github.com/informatics-isi-edu/deriva-py/releases",
   "asset_mappings": [
     {
@@ -49,6 +53,28 @@ DefaultConfig = {
 }
 
 
+class FileUploadState:
+    def __init__(self, state=UploadState.Pending, status="Pending", result=None):
+        self.state = state
+        self.status = status
+        self.result = result
+
+    def asdict(self):
+        return OrderedDict({
+            "State": self.state,
+            "Status": self.status,
+            "Result": self.result
+        })
+
+
+class UploadEntry(object):
+    def __init__(self, asset_group, asset_mapping, groupdict, path):
+        self.asset_group = asset_group
+        self.asset_mapping = asset_mapping
+        self.groupdict = groupdict
+        self.path = path
+
+
 class DerivaUpload(object):
     """
     Base class for upload tasks. Encapsulates a catalog instance and a hatrac store instance and provides some common
@@ -77,7 +103,7 @@ class DerivaUpload(object):
         self.metadata = dict()
         self.catalog_metadata = {"table_metadata": {}}
         self.processor_output = dict()
-
+        self.identity = dict()
         self.file_list = OrderedDict()
         self.file_status = OrderedDict()
         self.skipped_files = set()
@@ -85,11 +111,16 @@ class DerivaUpload(object):
         self.override_credential_file = credential_file
         self.server = self.getDefaultServer() if not server else server
         self.dcctx_cid = dcctx_cid if dcctx_cid else self.__class__.__name__
+        signal.signal(signal.SIGINT, self.interrupt_handler)
         self.initialize()
 
     def __del__(self):
         self.cleanupTransferState()
 
+    def interrupt_handler(self, signum, frame):
+        logger.info("Caught interrupt signal.")
+        self.cancel()
+
     def initialize(self, cleanup=False):
         info = "%s v%s [Python %s, %s]" % (
             self.__class__.__name__, vu.get_installed_version(VERSION),
@@ -124,6 +155,15 @@ class DerivaUpload(object):
             del self.store
         self.store = HatracStore(protocol, host, self.credentials, session_config=session_config)
 
+        # determine identity
+        if self.credentials:
+            try:
+                attributes = self.catalog.get_authn_session().json()
+                self.identity = attributes.get("client", self.identity)
+            except Exception as e:
+                # not a big deal since the credential token being used could be expired
+                logger.debug("Unable to determine user identity from existing credential (may be expired): %s" % e)
+
         # init dcctx cid to a default
         self.set_dcctx_cid(self.dcctx_cid)
 
@@ -133,6 +173,10 @@ class DerivaUpload(object):
              2. Sub-classes of this class to bundle their own default configuration files in an arbitrary location.
              3. The updating of already deployed configuration files if bundled internal defaults are newer.             
         """
+        if self.override_config_file and not os.path.isfile(self.override_config_file):
+            raise DerivaUploadConfigurationError(
+                "The configuration file %s could not be found." % self.override_config_file)
+
         config_file = self.override_config_file if self.override_config_file else None
         # 1. If we don't already have a valid (i.e., overridden) path to a config file...
         if not (config_file and os.path.isfile(config_file)):
@@ -193,6 +237,11 @@ class DerivaUpload(object):
         self.credentials = credentials
         self.catalog.set_credentials(self.credentials, host)
         self.store.set_credentials(self.credentials, host)
+        try:
+            attributes = self.catalog.get_authn_session().json()
+            self.identity = attributes.get("client", self.identity)
+        except Exception as e:
+            logger.warning("Unable to determine user identity: %s" % e)
 
     def setConfig(self, config_file):
         if not config_file:
@@ -202,9 +251,9 @@ class DerivaUpload(object):
         else:
             self._update_internal_config(read_config(config_file))
         if not self.isVersionCompatible():
-            raise DerivaRestoreError("Upload version incompatibility detected",
-                                     "Current version: [%s], required version(s): %s." %
-                                     (self.getVersion(), self.getVersionCompatibility()))
+            raise RuntimeError("Upload version incompatibility detected",
+                               "Current version: [%s], required version(s): %s." %
+                               (self.getVersion(), self.getVersionCompatibility()))
     @classmethod
     def getDefaultServer(cls):
         servers = cls.getServers()
@@ -258,7 +307,7 @@ class DerivaUpload(object):
         if not (os.path.isfile(src) and os.path.isfile(dst)):
             return False
 
-        # This comparison wont work with PyInstaller single-file bundles because the bundle is extracted to a temp dir
+        # This comparison won't work with PyInstaller single-file bundles because the bundle is extracted to a temp dir
         # and every timestamp for every file in the bundle is reset to the bundle extraction/creation time.
         if getattr(sys, 'frozen', False):
             prefix = os.path.sep + "_MEI"
@@ -284,6 +333,10 @@ class DerivaUpload(object):
     @staticmethod
     def getCatalogTable(asset_mapping, metadata_dict=None):
         schema_name, table_name = asset_mapping.get('target_table', [None, None])
+        # allow for template substitution in the schema/table name tuple
+        if (schema_name and table_name) and metadata_dict is not None:
+            schema_name = schema_name.format(**metadata_dict)
+            table_name = table_name.format(**metadata_dict)
         if not (schema_name and table_name):
             metadata_dict_lower = {k.lower(): v for k, v in metadata_dict.items()}
             schema_name = metadata_dict_lower.get("schema")
@@ -293,7 +346,7 @@ class DerivaUpload(object):
         return '%s:%s' % (urlquote(schema_name), urlquote(table_name))
 
     @staticmethod
-    def interpolateDict(src, dst, allowNone=False):
+    def interpolateDict(src, dst, allow_none_column_list=[]):
         if not (isinstance(src, dict) and isinstance(dst, dict)):
             raise ValueError("Invalid input parameter type(s): (src = %s, dst = %s), expected (dict, dict)" % (
                 type(src).__name__, type(dst).__name__))
@@ -307,26 +360,27 @@ class DerivaUpload(object):
         for k, v in dst.items():
             try:
                 value = v.format(**src)
+                if k in src:
+                    v_type = type(src[k])
+                    value = v_type(value)
             except KeyError:
                 value = v
                 if value:
                     if value.startswith('{') and value.endswith('}'):
                         value = None
             dst.update({k: value})
-        # remove all None valued entries in the dest, if disallowed
-        if not allowNone:
-            empty = [k for k, v in dst.items() if v is None]
-            for k in empty:
-                del dst[k]
+        # remove all None valued entries in the dest, if column is not explicitly allowed
+        dst = {k: v for k, v in dst.items() if v is not None or k in allow_none_column_list}
 
         return dst
 
     @staticmethod
-    def pruneDict(src, dst, stringify=True):
+    def pruneDict(src, dst, allow_none_column_list=[]):
         dst = dst.copy()
         for k in dst.keys():
             value = src.get(k)
-            dst[k] = str(value) if (stringify and value is not None) else value
+            if value is not None or (value is None and k in allow_none_column_list):
+                dst[k] = value
         return dst
 
     def getCurrentConfigFilePath(self):
@@ -370,8 +424,6 @@ class DerivaUpload(object):
             with io.open(updated_config_path, 'w', newline='\n', encoding='utf-8') as config:
                 remote_config_data = json.dumps(
                     remote_config, ensure_ascii=False, sort_keys=True, separators=(',', ': '), indent=2)
-                if IS_PY2 and isinstance(remote_config_data, str):
-                    remote_config_data = unicode(remote_config_data, 'utf-8')
                 config.write(remote_config_data)
             new_md5 = hu.compute_file_hashes(updated_config_path, hashes=['md5'])['md5'][0]
             if current_md5 != new_md5:
@@ -393,6 +445,22 @@ class DerivaUpload(object):
             result.append(item)
         return result
 
+    @staticmethod
+    def archive_preprocessing_enabled(asset_mapping):
+        if not asset_mapping:
+            return False
+        if asset_mapping.get("archive_preprocessing_enabled", False):
+            return True
+        pre_processors = asset_mapping.get("pre_processors", [])
+        for processor_config in pre_processors:
+            processor_name = processor_config[PROCESSOR_NAME_KEY]
+            processor_type = processor_config.get(PROCESSOR_TYPE_KEY)
+            processor_impl = find_processor(processor_name, processor_type, bypass_whitelist=True)
+            if issubclass(processor_impl, ArchiveProcessor):
+                asset_mapping["archive_preprocessing_enabled"] = True
+                return True
+        return False
+
     def validateFile(self, root, path, name):
         if self.config.get("relative_path_validation", False):
             file_path = os.path.normpath(os.path.join(os.path.relpath(path, root), name))
@@ -402,7 +470,12 @@ class DerivaUpload(object):
         if not asset_mapping:
             return None
 
-        return asset_group, asset_mapping, groupdict, os.path.abspath(os.path.normpath(os.path.join(path, name)))
+        if self.archive_preprocessing_enabled(asset_mapping):
+            final_path = os.path.abspath(os.path.normpath(groupdict.get("archive_path", path)))
+        else:
+            final_path = os.path.abspath(os.path.normpath(os.path.join(path, name)))
+
+        return UploadEntry(asset_group, asset_mapping, groupdict, final_path)
 
     def scanDirectory(self, root, abort_on_invalid_input=False, purge_state=False):
         """
@@ -414,7 +487,7 @@ class DerivaUpload(object):
         """
         root = os.path.abspath(root)
         if not os.path.isdir(root):
-            raise ValueError("Invalid directory specified: [%s]" % root)
+            raise FileNotFoundError("Invalid directory specified: [%s]" % root)
         self.loadTransferState(root, purge=purge_state)
 
         logger.info("Scanning files in directory [%s]..." % root)
@@ -424,30 +497,32 @@ class DerivaUpload(object):
                 if file_name.startswith(self.DefaultTransferStateBaseName):
                     continue
                 file_path = os.path.normpath(os.path.join(path, file_name))
-                file_entry = self.validateFile(root, path, file_name)
-                if not file_entry:
+                upload_entry = self.validateFile(root, path, file_name)
+                if not upload_entry:
                     logger.info("Skipping file: [%s] -- Invalid file type or directory location." % file_path)
                     self.skipped_files.add(file_path)
                     if abort_on_invalid_input:
-                        raise ValueError("Invalid input detected, aborting.")
+                        raise DerivaUploadError("Invalid input detected, aborting.")
                 else:
-                    asset_group = file_entry[0]
-                    group_list = file_list.get(asset_group, [])
-                    group_list.append(file_entry)
+                    asset_group = upload_entry.asset_group
+                    group_list = file_list.get(asset_group, {})
+                    group_list.update({upload_entry.path: upload_entry})
                     file_list[asset_group] = group_list
 
         # make sure that file entries in both self.file_list and self.file_status are ordered by the declared order of
         # the asset_mapping for the file
         for group in sorted(file_list.keys()):
             self.file_list[group] = file_list[group]
-            for file_entry in file_list[group]:
-                file_path = file_entry[3]
-                logger.info("Including file: [%s]." % file_path)
+            for upload_entry in file_list[group].values():
+                file_path = upload_entry.path
+                logger.info("Including %s: [%s]." %
+                            ("directory (for archive)" if self.archive_preprocessing_enabled(
+                                upload_entry.asset_mapping) else "file", file_path))
                 status = self.getTransferStateStatus(file_path)
                 if status:
-                    self.file_status[file_path] = FileUploadState(UploadState.Paused, status)._asdict()
+                    self.file_status[file_path] = FileUploadState(UploadState.Paused, status).asdict()
                 else:
-                    self.file_status[file_path] = FileUploadState(UploadState.Pending, "Pending")._asdict()
+                    self.file_status[file_path] = FileUploadState(UploadState.Pending, "Pending").asdict()
 
     def getAssetMapping(self, file_path):
         """
@@ -469,12 +544,20 @@ class DerivaUpload(object):
                     continue
                 groupdict.update(match.groupdict())
             if ext_pattern:
+                if self.archive_preprocessing_enabled(asset_type):
+                    logger.warning("The 'ext_pattern' parameter is not compatible when archive preprocessing "
+                                   "is enabled. Only input directories matching 'dir_pattern' are supported.")
+                    continue
                 match = re.search(ext_pattern, path, re.IGNORECASE)
                 if not match:
                     logger.debug("The ext_pattern \"%s\" failed to match the input path [%s]" % (ext_pattern, path))
                     continue
                 groupdict.update(match.groupdict())
             if file_pattern:
+                if self.archive_preprocessing_enabled(asset_type):
+                    logger.warning("The 'file_pattern' parameter is not compatible when archive preprocessing "
+                                   "is enabled. Only input directories matching 'dir_pattern' are supported.")
+                    continue
                 match = re.search(file_pattern, path)
                 if not match:
                     logger.debug("The file_pattern \"%s\" failed to match the input path [%s]" % (file_pattern, path))
@@ -486,57 +569,77 @@ class DerivaUpload(object):
         return None, None, None
 
     def uploadFiles(self, status_callback=None, file_callback=None):
-        self.catalog_model = self.catalog.getCatalogModel()
+        if not self.identity:
+            raise DerivaUploadAuthenticationError("Unable to determine user identity for %s. "
+                                                  "Please ensure that you are authenticated successfully." %
+                                                  self.server_url)
+        completed = 0
         for group, assets in self.file_list.items():
-            for asset_group_num, asset_mapping, groupdict, file_path in assets:
+            if self.cancelled:
+                break
+            for entry in assets.values():
                 if self.cancelled:
-                    self.file_status[file_path] = FileUploadState(UploadState.Cancelled, "Cancelled by user")._asdict()
-                    continue
+                    self.file_status[entry.path] = FileUploadState(UploadState.Cancelled, "Cancelled by user").asdict()
+                    break
                 try:
-                    self.file_status[file_path] = FileUploadState(UploadState.Running, "In-progress")._asdict()
+                    self.file_status[entry.path] = FileUploadState(UploadState.Running, "In-progress").asdict()
                     if status_callback:
                         status_callback()
-                    self.uploadFile(file_path, asset_mapping, groupdict, file_callback)
+                    result = self.uploadFile(entry.path,
+                                             entry.asset_mapping,
+                                             entry.groupdict,
+                                             file_callback or self.defaultFileCallback)
                     if self.cancelled:
-                        self.file_status[file_path] = FileUploadState(UploadState.Cancelled,
-                                                                      "Cancelled by user")._asdict()
+                        self.file_status[entry.path] = FileUploadState(UploadState.Cancelled,
+                                                                       "Cancelled by user").asdict()
+                        break
                     else:
-                        self.file_status[file_path] = FileUploadState(UploadState.Success, "Complete")._asdict()
+                        self.file_status[entry.path] = FileUploadState(UploadState.Success, "Complete", result).asdict()
+                        completed += 1
                 except HatracJobPaused:
-                    status = self.getTransferStateStatus(file_path)
+                    status = self.getTransferStateStatus(entry.path)
                     if status:
-                        self.file_status[file_path] = FileUploadState(
-                            UploadState.Paused, "Paused: %s" % status)._asdict()
+                        self.file_status[entry.path] = FileUploadState(
+                            UploadState.Paused, "Paused: %s" % status).asdict()
                     continue
                 except HatracJobTimeout:
-                    status = self.getTransferStateStatus(file_path)
+                    status = self.getTransferStateStatus(entry.path)
                     if status:
-                        self.file_status[file_path] = FileUploadState(UploadState.Timeout, "Timeout")._asdict()
+                        self.file_status[entry.path] = FileUploadState(UploadState.Timeout, "Timeout").asdict()
                     continue
                 except HatracJobAborted:
-                    self.file_status[file_path] = FileUploadState(UploadState.Aborted, "Aborted by user")._asdict()
+                    self.file_status[entry.path] = FileUploadState(UploadState.Aborted, "Aborted by user").asdict()
                 except:
                     logger.debug("Unexpected exception", exc_info=sys.exc_info())
                     (etype, value, traceback) = sys.exc_info()
-                    self.file_status[file_path] = FileUploadState(UploadState.Failed, format_exception(value))._asdict()
-                self.delTransferState(file_path)
+                    self.file_status[entry.path] = FileUploadState(UploadState.Failed, format_exception(value)).asdict()
+                self.delTransferState(entry.path)
                 if status_callback:
                     status_callback()
 
         failed_uploads = dict()
-        for key, value in self.file_status.items():
-            if (value["State"] == UploadState.Failed) or (value["State"] == UploadState.Timeout):
-                failed_uploads[key] = value["Status"]
+        try:
+            for key, value in self.file_status.items():
+                if (value["State"] == UploadState.Failed) or (value["State"] == UploadState.Timeout):
+                    failed_uploads[key] = value["Status"]
 
-        if self.skipped_files:
-            logger.warning("The following file(s) were skipped because they did not satisfy the matching criteria "
-                           "of the configuration:\n\n%s\n" % '\n'.join(sorted(self.skipped_files)))
+            if self.skipped_files:
+                logger.warning("The following %d file(s) were skipped because they did not satisfy the matching "
+                               "criteria of the configuration:\n\n%s\n" %
+                               (len(self.skipped_files), '\n'.join(sorted(self.skipped_files))))
 
-        if failed_uploads:
-            logger.warning("The following file(s) failed to upload due to errors:\n\n%s\n" %
-                           '\n'.join(["%s -- %s" % (key, failed_uploads[key])
-                                      for key in sorted(failed_uploads.keys())]))
-            raise RuntimeError("One or more file(s) failed to upload due to errors.")
+            if failed_uploads:
+                logger.warning("The following %d file(s) failed to upload due to errors:\n\n%s\n" %
+                               (len(failed_uploads), '\n'.join(["%s -- %s" % (key, failed_uploads[key])
+                                                                for key in sorted(failed_uploads.keys())])))
+                raise RuntimeError("%s file(s) failed to upload due to errors." % len(failed_uploads))
+        finally:
+            logger.info("File upload processing completed: %s files were uploaded successfully, "
+                        "%s files failed to upload due to errors, "
+                        "%s files were skipped because they did not satisfy the matching criteria of the configuration."
+                        % (completed, len(failed_uploads), len(self.skipped_files)))
+
+        return self.file_status
 
     def uploadFile(self, file_path, asset_mapping, match_groupdict, callback=None):
         """
@@ -547,22 +650,29 @@ class DerivaUpload(object):
         :param callback:
         :return:
         """
-        logger.info("Processing file: [%s]" % file_path)
+        logger.info("Processing: [%s]" % file_path)
 
         if asset_mapping.get("asset_type", "file") == "table":
-            self._uploadTable(file_path, asset_mapping, match_groupdict)
+            return self._uploadTable(file_path, asset_mapping, match_groupdict)
         else:
-            self._uploadAsset(file_path, asset_mapping, match_groupdict, callback)
+            return self._uploadAsset(file_path, asset_mapping, match_groupdict, callback)
 
     def _uploadAsset(self, file_path, asset_mapping, match_groupdict, callback=None):
 
         # 1. Populate initial file metadata from directory scan pattern matches
-        file_name = self.getFileDisplayName(file_path)
-        logger.info("Computing metadata for file: [%s]." % file_name)
         self._initFileMetadata(file_path, asset_mapping, match_groupdict)
 
-        # 2. Compute checksum(s) for current file and add to metadata
-        logger.info("Computing checksums for file: [%s]. Please wait..." % file_name)
+        # 2. Execute any configured preprocessors
+        self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=PRE_PROCESSORS_KEY)
+        if PROCESSOR_MODIFIED_FILE_PATH_KEY in self.processor_output:
+            file_path = self.processor_output[PROCESSOR_MODIFIED_FILE_PATH_KEY]
+        self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))
+        logger.info("Computed metadata for: [%s]." % file_path)
+        if logger.isEnabledFor(logging.DEBUG):
+            logger.debug("Current metadata: %s" % self.metadata)
+
+        # 3. Compute checksum(s) for current file and add to metadata
+        logger.info("Computing checksums for file: [%s]. Please wait..." % file_path)
         hashes = self.getFileHashes(file_path, asset_mapping.get('checksum_types', ['md5', 'sha256']))
         for alg, checksum in hashes.items():
             alg = alg.lower()
@@ -571,15 +681,14 @@ class DerivaUpload(object):
         if self.cancelled:
             return
 
-        # 3. Execute any configured preprocessors
-        self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=PRE_PROCESSORS_KEY)
-
         # 4. Populate additional metadata by querying the catalog
         self._queryFileMetadata(asset_mapping)
+        if logger.isEnabledFor(logging.DEBUG):
+            logger.debug("Updated metadata: %s" % self.metadata)
 
         # 5. If "create_record_before_upload" specified in asset_mapping, check for an existing record, creating a new
-        #    one if necessary. Otherwise delay this logic until after the file upload.
-        record = None
+        #    one if necessary. Otherwise, delay this logic until after the file upload.
+        result = record = None
         if stob(asset_mapping.get("create_record_before_upload", False)):
             record = self._getFileRecord(asset_mapping)
 
@@ -599,19 +708,26 @@ class DerivaUpload(object):
                                allow_versioning=stob(hatrac_options.get("allow_versioning", True)),
                                callback=callback)
         logger.debug("Hatrac upload successful. Result object URI: %s" % versioned_uri)
-        if stob(hatrac_options.get("versioned_uris", True)):
+        versioned_uris = True
+        if "versioned_uris" in hatrac_options:
+            versioned_uris = stob(hatrac_options.get("versioned_uris", True))
+        if "versioned_urls" in hatrac_options:
+            versioned_uris = stob(hatrac_options.get("versioned_urls", True))
+        if versioned_uris:
             self.metadata["URI"] = versioned_uri
         else:
             self.metadata["URI"] = versioned_uri.rsplit(":")[0]
-        self.metadata["URI_urlencoded"] = urlquote(self.metadata["URI"])
+        safe_overrides = asset_mapping.get("url_encoding_safe_overrides", {}).get("URI", "")
+        self.metadata["URI_urlencoded"] = urlquote(self.metadata["URI"], safe=safe_overrides)
 
         # 7. Check for an existing record and create a new one if necessary
         if not record:
-            record = self._getFileRecord(asset_mapping)
+            record, result = self._getFileRecord(asset_mapping)
 
         # 8. Update an existing record, if necessary
         column_map = asset_mapping.get("column_map", {})
-        updated_record = self.interpolateDict(self.metadata, column_map)
+        allow_none_col_list = asset_mapping.get("allow_empty_columns_on_update", [])
+        updated_record = self.interpolateDict(self.metadata, column_map, allow_none_col_list)
         if updated_record != record:
             record_update_template = asset_mapping.get("record_update_template")
             require_record_update_template = stob(asset_mapping.get("require_record_update_template", False))
@@ -620,11 +736,19 @@ class DerivaUpload(object):
                     "A required 'record_update_template' parameter for this asset mapping could not be found in the "
                     "configuration. The record will not be updated.")
             logger.info("Updating catalog for file [%s]" % self.getFileDisplayName(file_path))
-            self._catalogRecordUpdate(self.metadata['target_table'], record, updated_record, record_update_template)
+            result = self._catalogRecordUpdate(self.metadata['target_table'],
+                                               record,
+                                               updated_record,
+                                               record_update_template)[0]
+            if logger.isEnabledFor(logging.DEBUG):
+                logger.debug("Updated catalog for file [%s]: %s" % (self.getFileDisplayName(file_path), result))
+            record, result = self._getFileRecord(asset_mapping)
 
         # 9. Execute any configured post_processors
         self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=POST_PROCESSORS_KEY)
 
+        return result
+
     def _uploadTable(self, file_path, asset_mapping, match_groupdict, callback=None):
         if self.cancelled:
             return None
@@ -659,7 +783,9 @@ class DerivaUpload(object):
         Helper function that queries the catalog to get a record linked to the asset, or create it if it doesn't exist.
         :return: the file record
         """
+        record = None
         column_map = asset_mapping.get("column_map", {})
+        allow_none_col_list = asset_mapping.get("allow_empty_columns_on_update", [])
         rqt = asset_mapping['record_query_template']
         try:
             path = rqt.format(**self.metadata)
@@ -667,14 +793,16 @@ class DerivaUpload(object):
             raise DerivaUploadConfigurationError("Record query template substitution error: %s" % format_exception(e))
         result = self.catalog.get(path).json()
         if result:
-            self._updateFileMetadata(result[0], no_overwrite=True)
-            return self.pruneDict(result[0], column_map)
+            record = result[0]
+            self._updateFileMetadata(record, no_overwrite=True)
+            return self.pruneDict(record, column_map, allow_none_col_list), record
         else:
             row = self.interpolateDict(self.metadata, column_map)
             result = self._catalogRecordCreate(self.metadata['target_table'], row)
             if result:
-                self._updateFileMetadata(result[0])
-            return self.interpolateDict(self.metadata, column_map, allowNone=True)
+                record = result[0]
+                self._updateFileMetadata(record)
+            return self.interpolateDict(self.metadata, column_map, allow_none_column_list=allow_none_col_list), record
 
     def _urlEncodeMetadata(self, safe_overrides=None):
         urlencoded = dict()
@@ -689,16 +817,31 @@ class DerivaUpload(object):
     def _initFileMetadata(self, file_path, asset_mapping, match_groupdict):
         self.metadata.clear()
         self._updateFileMetadata(match_groupdict)
-
         self.metadata['target_table'] = self.getCatalogTable(asset_mapping, match_groupdict)
+
         self.metadata["file_name"] = self.getFileDisplayName(file_path)
         self.metadata["file_size"] = self.getFileSize(file_path)
+        if "file_ext" not in self.metadata:
+            self.metadata["file_ext"] = "".join(pathlib.PurePath(file_path).suffixes)
+        self.metadata["base_path"] = os.path.dirname(file_path)
+        self.metadata["base_name"] = self.metadata["file_name"].rsplit(
+            self.metadata["file_ext"])[0] if self.metadata["file_ext"] else self.metadata["file_name"]
+
+        time = datetime.datetime.now()
+        self.metadata["_upload_year_"] = time.year
+        self.metadata["_upload_month_"] = time.month
+        self.metadata["_upload_day_"] = time.day
+        self.metadata["_upload_time_"] = time.timestamp()
+        self.metadata["_identity_id"] = self.identity.get("id", "anonymous")
+        self.metadata["_identity_display_name"] = self.identity.get("display_name")
+        self.metadata["_identity_full_name"] = self.identity.get("full_name")
+        self.metadata["_identity_email"] = self.identity.get("email")
 
         self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))
 
     def _updateFileMetadata(self, src, strict=False, no_overwrite=False):
         if not (isinstance(src, dict)):
-            ValueError("Invalid input parameter type(s): (src = %s), expected (dict)" % type(src).__name__)
+            raise ValueError("Invalid input parameter type(s): (src = %s), expected (dict)" % type(src).__name__)
         dst = src.copy()
         for k in src.keys():
             if strict:
@@ -707,7 +850,7 @@ class DerivaUpload(object):
                                    "ignoring value: %s " % (k, src[k]))
                     del dst[k]
                     continue
-            # dont overwrite any existing metadata field
+            # don't overwrite any existing metadata field
             if no_overwrite:
                 if k in self.metadata:
                     del dst[k]
@@ -717,7 +860,10 @@ class DerivaUpload(object):
         """
         Helper function that queries the catalog to get required metadata for a given file/asset
         """
-        for uri in asset_mapping.get("metadata_query_templates", []):
+        metadata_queries = asset_mapping.get("metadata_query_templates", [])
+        if logger.isEnabledFor(logging.DEBUG) and metadata_queries:
+            logger.debug("Querying catalog for additional metadata...")
+        for uri in metadata_queries:
             try:
                 path = uri.format(**self.metadata)
             except KeyError as e:
@@ -725,7 +871,6 @@ class DerivaUpload(object):
             result = self.catalog.get(path).json()
             if result:
                 self._updateFileMetadata(result[0], True)
-                self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))
             else:
                 raise RuntimeError("Metadata query did not return any results: %s" % path)
 
@@ -737,6 +882,7 @@ class DerivaUpload(object):
             except KeyError as e:
                 logger.warning("Column value template substitution error: %s" % format_exception(e))
                 continue
+
         self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))
 
     def _getFileExtensionMetadata(self, ext):
@@ -748,12 +894,14 @@ class DerivaUpload(object):
     def _getFileHatracMetadata(self, asset_mapping):
         try:
             hatrac_templates = asset_mapping["hatrac_templates"]
+            # convert None values to empty strings for URI and content-disposition template replacement
+            metadata = {k: ('' if v is None else v) for k, v in self.metadata.items()}
             # URI is required
-            self.metadata["URI"] = hatrac_templates["hatrac_uri"].format(**self.metadata)
+            self.metadata["URI"] = hatrac_templates["hatrac_uri"].format(**metadata)
             # overridden content-disposition is optional
             content_disposition = hatrac_templates.get("content-disposition")
             if content_disposition:
-                filename = content_disposition.format(**self.metadata)
+                filename = content_disposition.format(**metadata)
             else:
                 filename = urlparse(self.metadata["URI"]).path.rsplit("/", 1)[-1]
 
@@ -823,8 +971,7 @@ class DerivaUpload(object):
                                            cancel_job_on_error=False)
             return self.store.finalize_upload_job(path, job_id)
         else:
-            logger.info("Uploading file: [%s] to host %s. Please wait..." % (
-                self.getFileDisplayName(file_path), self.server_url))
+            logger.info("Uploading file: [%s] to host %s. Please wait..." % (file_path, self.server_url))
             return self.store.put_loc(uri,
                                       file_path,
                                       md5=md5,
@@ -853,6 +1000,9 @@ class DerivaUpload(object):
 
     def _validate_row_key_constraints(self, catalog_table, row):
         logger.debug("Validating row key constraints for %s: %s" % (catalog_table, row))
+        if not self.catalog_model:
+            logger.debug("Fetching catalog model...")
+            self.catalog_model = self.catalog.getCatalogModel()
         schema_name, table_name = self.catalog.splitQualifiedCatalogName(catalog_table)
         schema = self.catalog_model.schemas.get(schema_name)
         table = schema.tables.get(table_name)
@@ -977,6 +1127,9 @@ class DerivaUpload(object):
                         **kwargs)
                     proc_class = processor.__class__.__module__
                     proc_name = processor.__class__.__name__
+                    if processor_params is not None and processor_params.get(
+                            PROCESSOR_REQUIRES_METADATA_QUERY_KEY, False):
+                        self._queryFileMetadata(asset_mapping)
                     logger.debug("Attempting to execute upload processor class %s from module: %s" %
                                  (proc_name, proc_class))
                     output = processor.process()
@@ -1021,8 +1174,8 @@ class DerivaUpload(object):
         for root, dirs, files in walk(path, followlinks=True):
             if filename in files:
                 # found a file
-                found_path = os.path.join(root, filename)
-                file_paths.add(os.path.realpath(found_path))
+                found_path = os.path.normcase(os.path.join(root, filename))
+                file_paths.add(os.path.normcase(os.path.realpath(found_path)))
                 continue
 
         # Next, ascend from the base path looking for the same filename in all parent dirs
@@ -1033,7 +1186,7 @@ class DerivaUpload(object):
                 break
             for entry in scandir(parent):
                 if (entry.name == filename) and entry.is_file():
-                    file_paths.add(os.path.realpath(entry.path))
+                    file_paths.add(os.path.normcase(os.path.realpath(os.path.normcase(entry.path))))
             current = parent
 
         return file_paths
@@ -1059,13 +1212,13 @@ class DerivaUpload(object):
                                         "%s" % (os.path.dirname(path), format_exception(e)))
 
     def loadTransferState(self, directory, purge=False):
-        transfer_state_file_path = os.path.join(directory, self.getTransferStateFileName())
+        transfer_state_file_path = os.path.normcase(os.path.join(directory, self.getTransferStateFileName()))
         if purge:
             self.delete_dependent_locks(directory)
         self.acquire_dependent_locks(directory)
         try:
             if not os.path.isfile(transfer_state_file_path):
-                with lock_file(transfer_state_file_path, mode="wb" if IS_PY2 else "w") as tsfp:
+                with lock_file(transfer_state_file_path, "w") as tsfp:
                     json.dump(self.transfer_state, tsfp)
 
             transfer_state_lock = self.transfer_state_locks.get(transfer_state_file_path)
