@@ -1,8 +1,9 @@
 import os
 import sys
+import json
 import traceback
 from deriva.transfer import DerivaUpload, DerivaUploadError, DerivaUploadConfigurationError, \
-    DerivaUploadCatalogCreateError, DerivaUploadCatalogUpdateError
+    DerivaUploadCatalogCreateError, DerivaUploadCatalogUpdateError, DerivaUploadAuthenticationError
 from deriva.core import BaseCLI, write_config, format_credential, format_exception, urlparse
 
 
@@ -17,6 +18,12 @@ class DerivaUploadCLI(BaseCLI):
         self.parser.add_argument('--purge-state', action="store_true",
                                  help="Purge (delete) any existing transfer state files found in the directory "
                                       "hierarchy of the input path.")
+        self.parser.add_argument('--dry-run', action="store_true",
+                                 help="Scan the input directory for matching files to upload, "
+                                      "but do not transfer any files.")
+        self.parser.add_argument('--output-file', metavar='<file>',
+                                 help="Optional path where a JSON-formatted output file will be written, "
+                                      "containing file upload status and associated metadata.")
         self.parser.add_argument("--catalog", default=1, metavar="<1>", help="Catalog number. Default: 1")
         self.parser.add_argument("path", metavar="<input dir>", help="Path to an input directory.")
         self.uploader = uploader
@@ -30,7 +37,9 @@ class DerivaUploadCLI(BaseCLI):
                config_file=None,
                credential_file=None,
                no_update=False,
-               purge=False):
+               purge=False,
+               dry_run=False,
+               output_file=None):
 
         if not issubclass(uploader, DerivaUpload):
             raise TypeError("DerivaUpload subclass required")
@@ -57,7 +66,11 @@ class DerivaUploadCLI(BaseCLI):
             raise RuntimeError("Version incompatibility detected", "Current version: [%s], required version(s): %s." % (
                 deriva_uploader.getVersion(), deriva_uploader.getVersionCompatibility()))
         deriva_uploader.scanDirectory(data_path, abort_on_invalid_input=False, purge_state=purge)
-        deriva_uploader.uploadFiles(file_callback=deriva_uploader.defaultFileCallback)
+        if not dry_run:
+            results = deriva_uploader.uploadFiles()
+            if output_file:
+                with open(output_file, "w") as output:
+                    json.dump(results, output)
         deriva_uploader.cleanup()
 
     def main(self):
@@ -78,9 +91,11 @@ class DerivaUploadCLI(BaseCLI):
                                    args.config_file,
                                    args.credential_file,
                                    args.no_config_update,
-                                   args.purge_state)
-        except (RuntimeError, DerivaUploadError, DerivaUploadConfigurationError, DerivaUploadCatalogCreateError,
-                DerivaUploadCatalogUpdateError) as e:
+                                   args.purge_state,
+                                   args.dry_run,
+                                   args.output_file)
+        except (RuntimeError, FileNotFoundError, DerivaUploadError, DerivaUploadConfigurationError,
+                DerivaUploadCatalogCreateError, DerivaUploadCatalogUpdateError, DerivaUploadAuthenticationError) as e:
             sys.stderr.write(("\n" if not args.quiet else "") + format_exception(e))
             if args.debug:
                 traceback.print_exc()
