@@ -10,21 +10,8 @@ import logging
 from operator import itemgetter
 import os
 import unittest
-import sys
 from deriva.core import DerivaServer, get_credential, ermrest_model as _em, __version__
-from deriva.core.datapath import DataPathException, Min, Max, Sum, Avg, Cnt, CntD, Array, ArrayD, Bin
-
-# unittests did not support 'subTests' until 3.4
-if sys.version_info[0] < 3 or sys.version_info[1] < 4:
-    HAS_SUBTESTS = False
-else:
-    HAS_SUBTESTS = True
-
-# unittests did not support 'assertWarns' until 3.2
-if sys.version_info[0] < 3 or sys.version_info[1] < 2:
-    HAS_ASSERTWARNS = False
-else:
-    HAS_ASSERTWARNS = True
+from deriva.core.datapath import *
 
 try:
     from pandas import DataFrame
@@ -87,33 +74,33 @@ def define_test_schema(catalog):
     )
     isa.create_table(table_def)
 
-    # create TNAME_EXPERIMENT table
-    table_def = _em.Table.define(
-        TNAME_EXPERIMENT,
-        column_defs=[
-            _em.Column.define(cname, ctype) for (cname, ctype) in [
-                ('Name', _em.builtin_types.text),
-                ('Amount', _em.builtin_types.int4),
-                ('Time', _em.builtin_types.timestamptz),
-                ('Type', _em.builtin_types.text),
-                ('Project Investigator', _em.builtin_types.text),
-                ('Project_Num', _em.builtin_types.int4),
-                ('Empty', _em.builtin_types.int4)
+    # experiment table definition helper
+    def exp_table_def(exp_table_name):
+        return _em.Table.define(
+            exp_table_name,
+            column_defs=[
+                _em.Column.define(cname, ctype) for (cname, ctype) in [
+                    ('Name', _em.builtin_types.text),
+                    ('Amount', _em.builtin_types.int4),
+                    ('Time', _em.builtin_types.timestamptz),
+                    ('Type', _em.builtin_types.text),
+                    ('Project Investigator', _em.builtin_types.text),
+                    ('Project_Num', _em.builtin_types.int4),
+                    ('Empty', _em.builtin_types.int4)
+                ]
+            ],
+            key_defs=[
+                _em.Key.define(['Name'])
+            ],
+            fkey_defs=[
+                _em.ForeignKey.define(['Type'], SNAME_VOCAB, TNAME_EXPERIMENT_TYPE, ['ID']),
+                _em.ForeignKey.define(['Project Investigator', 'Project_Num'], SNAME_ISA, TNAME_PROJECT, ['Investigator', 'Num'])
             ]
-        ],
-        key_defs=[
-            _em.Key.define(['Name'])
-        ],
-        fkey_defs=[
-            _em.ForeignKey.define(['Type'], SNAME_VOCAB, TNAME_EXPERIMENT_TYPE, ['ID']),
-            _em.ForeignKey.define(['Project Investigator', 'Project_Num'], SNAME_ISA, TNAME_PROJECT, ['Investigator', 'Num'])
-        ]
-    )
-    isa.create_table(table_def)
+        )
 
-    # create copy of TNAME_EXPERIMENT table
-    table_def['table_name'] = TNAME_EXPERIMENT_COPY
-    isa.create_table(table_def)
+    # create experiment tables
+    isa.create_table(exp_table_def(TNAME_EXPERIMENT))
+    isa.create_table(exp_table_def(TNAME_EXPERIMENT_COPY))
 
 
 def _generate_experiment_entities(types, count):
@@ -141,11 +128,12 @@ def populate_test_catalog(catalog):
     """Populate the test catalog."""
     paths = catalog.getPathBuilder()
     logger.debug("Inserting project...")
-    logger.debug("Inserting experiment types...")
     proj_table = paths.schemas[SNAME_ISA].tables[TNAME_PROJECT]
+    logger.debug("Inserting investigators...")
     proj_table.insert([
         {"Investigator": TEST_PROJ_INVESTIGATOR, "Num": TEST_PROJ_NUM}
     ])
+    logger.debug("Inserting experiment types...")
     type_table = paths.schemas[SNAME_VOCAB].tables[TNAME_EXPERIMENT_TYPE]
     types = type_table.insert([
         {"Name": "{}".format(name), "Description": "NA"} for name in range(TEST_EXPTYPE_MAX)
@@ -347,7 +335,6 @@ class DatapathTests (unittest.TestCase):
                 Min(self.experiment.column_definitions['Amount'])
             )
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
     def test_aggregate_fns(self):
         tests = [
             ('min_amount',      Min,    0),
@@ -378,7 +365,6 @@ class DatapathTests (unittest.TestCase):
         self.assertIn('max_amount', result)
         self.assertEqual(result['max_amount'], TEST_EXP_MAX-1)
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
     def test_aggregate_fns_array_star(self):
         path = self.experiment.path
         tests = [
@@ -395,7 +381,6 @@ class DatapathTests (unittest.TestCase):
                 self.assertEqual(len(result['arr']), TEST_EXP_MAX)
                 self.assertIn('Time', result['arr'][0])
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
     def test_aggregate_fns_cnt_star(self):
         path = self.experiment.path
         tests = [
@@ -409,7 +394,6 @@ class DatapathTests (unittest.TestCase):
                 self.assertIn('cnt', result)
                 self.assertEqual(result['cnt'], TEST_EXP_MAX)
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
     def test_attributegroup_fns(self):
         tests = [
             ('one group key',     [self.experiment.column_definitions['Type']]),
@@ -420,7 +404,6 @@ class DatapathTests (unittest.TestCase):
             with self.subTest(name=test_name):
                 self._do_attributegroup_fn_subtests(group_key)
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
     def _do_attributegroup_fn_subtests(self, group_key):
         """Helper method for running common attributegroup subtests for different group keys."""
         tests = [
@@ -444,7 +427,6 @@ class DatapathTests (unittest.TestCase):
                 self.assertIn(name, result)
                 self.assertEqual(result[name], value)
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
     def test_attributegroup_w_bin(self):
         tests = [
             ('min/max given',     0,    TEST_EXP_MAX),
@@ -488,7 +470,6 @@ class DatapathTests (unittest.TestCase):
                         continue
                     self.assertTrue(compare(result[name], bin))
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
     def test_attributegroup_w_bin_sort(self):
         bin_name = 'bin'
         nbins = int(TEST_EXP_MAX/20)
@@ -515,7 +496,6 @@ class DatapathTests (unittest.TestCase):
                 self.assertIn(name, results[0])
                 self.assertTrue(compfn(name, results[0], results[1]))
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This test is not available unless running python 3.4+")
     def test_attributegroup_w_bin_resolution(self):
         binkey = self.experiment.column_definitions['Empty']
         binname = 'bin'
@@ -593,10 +573,28 @@ class DatapathTests (unittest.TestCase):
                 self.assertEqual(expected_results_len, len(results))
 
     def test_filter_equality(self):
+        # test with value
         results = self.experiment.filter(
             self.experiment.column_definitions['Name'] == TEST_EXP_NAME_FORMAT.format(1)
         ).entities()
         self.assertEqual(len(results), 1)
+        # test again with null
+        results = self.experiment.filter(
+            self.experiment.column_definitions['Name'] == None
+        ).entities()
+        self.assertEqual(len(results), 0)
+
+    def test_filter_notequal(self):
+        # test with value
+        results = self.experiment.filter(
+            self.experiment.column_definitions['Amount'] != 0
+        ).entities()
+        self.assertEqual(len(results), TEST_EXP_MAX-1)
+        # test again with null
+        results = self.experiment.filter(
+            self.experiment.column_definitions['Amount'] != None
+        ).entities()
+        self.assertEqual(len(results), TEST_EXP_MAX)
 
     def test_filter_inequality(self):
         results = self.experiment.filter(
@@ -610,6 +608,24 @@ class DatapathTests (unittest.TestCase):
         ).entities()
         self.assertEqual(len(results), 1)
 
+    @unittest.skipUnless(TEST_EXP_MAX >= 3, "this test was designed for at least 3 elements in the test set")
+    def test_fitler_w_quantifier(self):
+        int_args = [0, int(TEST_EXP_MAX/2), TEST_EXP_MAX-1, TEST_EXP_MAX*2]
+        str_args = [TEST_EXP_NAME_FORMAT.format(i) for i in int_args]
+        for name, cname, Q, args, expectation in [
+            ('any_3_ints', 'Amount', Any, int_args, 3),
+            ('all_1_int', 'Amount', All, [int_args[0]], 1),
+            ('all_3_ints', 'Amount', All, int_args, 0),
+            ('any_3_strs', 'Name', Any, str_args, 3),
+            ('all_1_str', 'Name', All, [str_args[0]], 1),
+            ('all_3_strs', 'Name', All, str_args, 0)
+        ]:
+            with self.subTest(name=name):
+                results = self.experiment.filter(
+                    self.experiment.column_definitions[cname] == Q(*args)
+                ).entities()
+                self.assertEqual(len(results), expectation)
+
     def test_filter_negation(self):
         results = self.experiment.filter(
             ~ (self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:]))
@@ -683,6 +699,13 @@ class DatapathTests (unittest.TestCase):
         df = DataFrame(results)
         self.assertEqual(len(df), TEST_EXP_MAX)
 
+    def test_insert_double_fetch(self):
+        entities = _generate_experiment_entities(self.types, 2)
+        results = self.experiment_copy.insert(entities)
+        rows1 = results.fetch()
+        rows2 = results.fetch()
+        self.assertEqual(rows1, rows2)
+        
     def test_insert_empty_entities(self):
         results = self.experiment_copy.insert(None)
         self.assertEqual(len(results), 0)
@@ -746,6 +769,25 @@ class DatapathTests (unittest.TestCase):
         with self.assertRaises(TypeError):
             self.experiment_type.update('this is not a dict')
 
+    def test_delete_whole_path(self):
+        self.experiment_copy.insert(_generate_experiment_entities(self.types, 10))
+        self.assertEqual(len(self.experiment_copy.entities()), 10)
+        self.experiment_copy.path.delete()
+        self.assertEqual(len(self.experiment_copy.entities()), 0)
+
+    def test_delete_filtered_path(self):
+        self.experiment_copy.insert(_generate_experiment_entities(self.types, 10))
+        expression = self.experiment_copy.column_definitions['Name'] == TEST_EXP_NAME_FORMAT.format(1)
+        self.assertEqual(len(self.experiment_copy.filter(expression).entities()), 1)
+        self.experiment_copy.filter(expression).delete()
+        self.assertEqual(len(self.experiment_copy.filter(expression).entities()), 0)
+
+    def test_delete_whole_table(self):
+        self.experiment_copy.insert(_generate_experiment_entities(self.types, 10))
+        self.assertEqual(len(self.experiment_copy.entities()), 10)
+        self.experiment_copy.delete()
+        self.assertEqual(len(self.experiment_copy.entities()), 0)
+
     def test_nondefaults(self):
         nondefaults = {'RID', 'RCB', 'RCT'}
         results = self.experiment.entities()
@@ -766,7 +808,6 @@ class DatapathTests (unittest.TestCase):
         for i in range(TEST_EXP_MAX):
             self.assertEqual(ig(results[i]), ig(entities_copy[i]), 'copied values do not match')
 
-    @unittest.skipUnless(HAS_SUBTESTS, "This test is not available unless running python 3.4+")
     def test_deepcopy_of_paths(self):
         paths = [
             self.experiment.path,
@@ -827,6 +868,26 @@ class DatapathTests (unittest.TestCase):
         self.assertEqual(path.context._name, path3.context._name, "Context of composed paths should equal far right-hand path's context")
         self.assertGreater(len(path.Experiment.entities()), 0, "Should have returned results")
 
+    def test_simple_denormalization(self):
+        entities = self.experiment.entities()
+        results = self.experiment.denormalize()
+        self.assertEqual(len(entities), len(results))
+        self.assertNotEqual(entities[0].keys(), results[0].keys())
+        self.assertIn('Type', results[0])
+        self.assertTrue(entities[0]['Type'].startswith('TEST:'))
+        self.assertTrue(results[0]['Type'])
+        self.assertFalse(results[0]['Type'].startswith('TEST:'))
+
+    def test_simple_denormalization_w_entities(self):
+        entities = self.experiment.entities()
+        results = self.experiment.denormalize(heuristic=simple_denormalization_with_whole_entities)
+        self.assertEqual(len(entities), len(results))
+        self.assertLess(len(entities[0].keys()), len(results[0].keys()))
+        self.assertIn('Experiment_Project Investigator_Project_Num_fkey', results[0])
+        self.assertIsInstance(results[0]['Experiment_Project Investigator_Project_Num_fkey'], list)
+        self.assertIsInstance(results[0]['Experiment_Project Investigator_Project_Num_fkey'][0], dict)
+        self.assertIn('RID', results[0]['Experiment_Project Investigator_Project_Num_fkey'][0])
+
 
 if __name__ == '__main__':
-    sys.exit(unittest.main())
+    unittest.main()
